import { EditorView } from '@codemirror/next/view';
import { SelectionRange } from '@codemirror/next/state';
import { codePointAt, minPairCodePoint, isWordChar, fromCodePoint } from '@codemirror/next/text';
import { keyName } from 'w3c-keyname';

const defaults = {
    brackets: ["(", "[", "{", "'", '"'],
    before: ")]}'\":;>"
};
/// Extension to enable bracket-closing behavior. When a closeable
/// bracket is typed, its closing bracket is immediately inserted
/// after the cursor. When closing a bracket directly in front of that
/// closing bracket, the cursor moves over the existing bracket. When
/// backspacing in between brackets, both are removed.
const closeBrackets = EditorView.domEventHandlers({ keydown });
const definedClosing = "()[]{}<>";
function closing(ch) {
    for (let i = 0; i < definedClosing.length; i += 2)
        if (definedClosing.charCodeAt(i) == ch)
            return definedClosing.charAt(i + 1);
    return fromCodePoint(ch < 128 ? ch : ch + 1);
}
function config(state, pos) {
    return state.languageDataAt("closeBrackets", pos)[0] || defaults;
}
function keydown(event, view) {
    if (event.ctrlKey || event.metaKey)
        return false;
    if (event.keyCode == 8) { // Backspace
        let tr = handleBackspace(view.state);
        if (!tr)
            return false;
        view.dispatch(tr);
        return true;
    }
    let key = keyName(event);
    if (key.length > 2 || key.length == 2 && codePointAt(key, 0) < minPairCodePoint)
        return false;
    let tr = handleInsertion(view.state, key);
    if (!tr)
        return false;
    view.dispatch(tr);
    return true;
}
/// Function that implements the extension's backspace behavior.
/// Exported mostly for testing purposes.
function handleBackspace(state) {
    let conf = config(state, state.selection.primary.head);
    let tokens = conf.brackets || defaults.brackets;
    let tr = state.t(), dont = null;
    tr.forEachRange(range => {
        if (!range.empty)
            return dont = range;
        let before = prevChar(state.doc, range.head);
        for (let token of tokens) {
            if (token == before && nextChar(state.doc, range.head) == closing(codePointAt(token, 0))) {
                let ref = tr.mapRef();
                tr.replace(range.head - token.length, range.head + token.length, "");
                return new SelectionRange(ref.mapPos(range.head));
            }
        }
        return dont = range;
    });
    return dont ? null : tr.scrollIntoView();
}
/// Implements the extension's behavior on text insertion. Again,
/// exported mostly for testing.
function handleInsertion(state, ch) {
    let conf = config(state, state.selection.primary.head);
    let tokens = conf.brackets || defaults.brackets;
    for (let tok of tokens) {
        let closed = closing(codePointAt(tok, 0));
        if (ch == tok)
            return closed == tok ? handleSame(state, tok, tokens.indexOf(tok + tok + tok) > -1)
                : handleOpen(state, tok, closed, conf.before || defaults.before);
        if (ch == closed)
            return handleClose(state, tok, closed);
    }
    return null;
}
function nextChar(doc, pos) {
    let next = doc.slice(pos, pos + 2);
    return next.length == 2 && codePointAt(next, 0) < minPairCodePoint ? next.slice(0, 1) : next;
}
function prevChar(doc, pos) {
    let prev = doc.slice(pos - 2, pos);
    return prev.length == 2 && codePointAt(prev, 0) < minPairCodePoint ? prev.slice(1) : prev;
}
function handleOpen(state, open, close, closeBefore) {
    let tr = state.t(), dont = null;
    tr.forEachRange(range => {
        if (!range.empty) {
            let ref = tr.mapRef(), dir = range.head - range.anchor;
            tr.replace(range.to, range.to, close);
            tr.replace(ref.mapPos(range.from), ref.mapPos(range.from), open);
            return new SelectionRange(ref.mapPos(range.anchor, dir), ref.mapPos(range.head, -dir));
        }
        let next = nextChar(state.doc, range.head);
        if (!next || /\s/.test(next) || closeBefore.indexOf(next) > -1) {
            tr.replace(range.head, range.head, open + close);
            return new SelectionRange(Math.min(tr.doc.length, range.head + open.length));
        }
        return dont = range;
    });
    return dont ? null : tr.scrollIntoView();
}
function handleClose(state, _open, close) {
    let tr = state.t(), dont = null;
    tr.forEachRange(range => {
        if (range.empty && close == nextChar(state.doc, range.head))
            return new SelectionRange(range.head + close.length);
        return dont = range;
    });
    return dont ? null : tr.scrollIntoView();
}
// Handles cases where the open and close token are the same, and
// possibly triple quotes (as in `"""abc"""`-style quoting).
function handleSame(state, token, allowTriple) {
    let tr = state.t(), dont = null;
    tr.forEachRange(range => {
        if (!range.empty) {
            let ref = tr.mapRef(), dir = range.head - range.anchor;
            tr.replace(range.to, range.to, token);
            tr.replace(ref.mapPos(range.from), ref.mapPos(range.from), token);
            return new SelectionRange(ref.mapPos(range.anchor, dir), ref.mapPos(range.head, -dir));
        }
        let pos = range.head, next = nextChar(state.doc, pos);
        if (next == token) {
            if (nodeStart(state, pos)) {
                tr.replace(pos, pos, token + token);
                return new SelectionRange(Math.min(tr.doc.length, pos + token.length));
            }
            else {
                let isTriple = allowTriple && state.doc.slice(pos, pos + token.length * 3) == token + token + token;
                return new SelectionRange(pos + token.length * (isTriple ? 3 : 1));
            }
        }
        else if (allowTriple && state.doc.slice(pos - 2 * token.length, pos) == token + token &&
            nodeStart(state, pos - 2 * token.length)) {
            tr.replace(pos, pos, token + token + token + token);
            return new SelectionRange(Math.min(tr.doc.length, pos + token.length));
        }
        else if (!isWordChar(next)) {
            let prev = state.doc.slice(pos - 1, pos);
            if (!isWordChar(prev) && prev != token) {
                tr.replace(pos, pos, token + token);
                return new SelectionRange(Math.min(tr.doc.length, pos + token.length));
            }
        }
        return dont = range;
    });
    return dont ? null : tr.scrollIntoView();
}
function nodeStart(state, pos) {
    let tree = state.tree.resolve(pos + 1);
    return tree.parent && tree.start == pos;
}

export { closeBrackets, handleBackspace, handleInsertion };
