import { EditorSelection, IndentContext, SelectionRange, Transaction, EditorState } from '@codemirror/next/state';

function moveSelection(view, dir, granularity) {
    let transaction = view.state.t().forEachRange(range => {
        if (!range.empty && granularity != "lineboundary")
            return new SelectionRange(dir == "left" || dir == "backward" ? range.from : range.to);
        return new SelectionRange(view.movePos(range.head, dir, granularity, "move"));
    });
    if (transaction.selection.eq(view.state.selection))
        return false;
    if (granularity == "line")
        transaction.annotate(Transaction.preserveGoalColumn, true);
    view.dispatch(transaction.scrollIntoView());
    return true;
}
/// Move the selection one character to the left (which is backward in
/// left-to-right text, forward in right-to-left text).
const moveCharLeft = view => moveSelection(view, "left", "character");
/// Move the selection one character to the right.
const moveCharRight = view => moveSelection(view, "right", "character");
/// Move the selection one word to the left.
const moveWordLeft = view => moveSelection(view, "left", "word");
/// Move the selection one word to the right.
const moveWordRight = view => moveSelection(view, "right", "word");
/// Move the selection one line up.
const moveLineUp = view => moveSelection(view, "backward", "line");
/// Move the selection one line down.
const moveLineDown = view => moveSelection(view, "forward", "line");
/// Move the selection to the start of the line.
const moveLineStart = view => moveSelection(view, "backward", "lineboundary");
/// Move the selection to the end of the line.
const moveLineEnd = view => moveSelection(view, "forward", "lineboundary");
function extendSelection(view, dir, granularity) {
    let transaction = view.state.t().forEachRange(range => {
        return new SelectionRange(range.anchor, view.movePos(range.head, dir, granularity, "extend"));
    });
    if (transaction.selection.eq(view.state.selection))
        return false;
    if (granularity == "line")
        transaction.annotate(Transaction.preserveGoalColumn, true);
    view.dispatch(transaction.scrollIntoView());
    return true;
}
/// Move the selection head one character to the left, while leaving
/// the anchor in place.
const extendCharLeft = view => extendSelection(view, "left", "character");
/// Move the selection head one character to the right.
const extendCharRight = view => extendSelection(view, "right", "character");
/// Move the selection head one word to the left.
const extendWordLeft = view => extendSelection(view, "left", "word");
/// Move the selection head one word to the right.
const extendWordRight = view => extendSelection(view, "right", "word");
/// Move the selection head one line up.
const extendLineUp = view => extendSelection(view, "backward", "line");
/// Move the selection head one line down.
const extendLineDown = view => extendSelection(view, "forward", "line");
/// Move the selection head to the start of the line.
const extendLineStart = view => extendSelection(view, "backward", "lineboundary");
/// Move the selection head to the end of the line.
const extendLineEnd = view => extendSelection(view, "forward", "lineboundary");
/// Move the selection to the start of the document.
const selectDocStart = ({ state, dispatch }) => {
    dispatch(state.t().setSelection(EditorSelection.single(0)).scrollIntoView());
    return true;
};
/// Move the selection to the end of the document.
const selectDocEnd = ({ state, dispatch }) => {
    dispatch(state.t().setSelection(EditorSelection.single(state.doc.length)).scrollIntoView());
    return true;
};
/// Select the entire document.
const selectAll = ({ state, dispatch }) => {
    dispatch(state.t().setSelection(EditorSelection.single(0, state.doc.length)));
    return true;
};
function deleteText(view, dir) {
    let transaction = view.state.t().forEachRange((range, transaction) => {
        let { from, to } = range;
        if (from == to) {
            let target = view.movePos(range.head, dir, "character", "move");
            from = Math.min(from, target);
            to = Math.max(to, target);
        }
        if (from == to)
            return range;
        transaction.replace(from, to, "");
        return new SelectionRange(from);
    });
    if (!transaction.docChanged)
        return false;
    view.dispatch(transaction.scrollIntoView());
    return true;
}
/// Delete the character before the cursor (which is the one to left
/// in left-to-right text, but the one to the right in right-to-left
/// text).
const deleteCharBackward = view => deleteText(view, "backward");
/// Delete the character after the cursor.
const deleteCharForward = view => deleteText(view, "forward");
// FIXME support indenting by tab
function space(n) {
    let result = "";
    for (let i = 0; i < n; i++)
        result += " ";
    return result;
}
function getIndentation(cx, pos) {
    for (let f of cx.state.facet(EditorState.indentation)) {
        let result = f(cx, pos);
        if (result > -1)
            return result;
    }
    return -1;
}
/// Replace the selection with a newline and indent the newly created
/// line(s).
const insertNewlineAndIndent = ({ state, dispatch }) => {
    let i = 0, indentation = state.selection.ranges.map(r => {
        let indent = getIndentation(new IndentContext(state), r.from);
        return indent > -1 ? indent : /^\s*/.exec(state.doc.lineAt(r.from).slice(0, 50))[0].length;
    });
    dispatch(state.t().forEachRange(({ from, to }, tr) => {
        let indent = indentation[i++], line = tr.doc.lineAt(to);
        while (to < line.end && /s/.test(line.slice(to - line.start, to + 1 - line.start)))
            to++;
        let ref = tr.mapRef();
        tr.replace(from, to, ["", space(indent)]);
        return new SelectionRange(ref.mapPos(from, 1));
    }).scrollIntoView());
    return true;
};
/// Auto-indent the selected lines. This uses the [indentation
/// behavor](#state.EditorState^indentation) as source.
const indentSelection = ({ state, dispatch }) => {
    let lastLine = -1, positions = [];
    let updated = Object.create(null);
    let context = new IndentContext(state, start => {
        let found = updated[start];
        return found == null ? -1 : found;
    });
    for (let range of state.selection.ranges) {
        for (let { start, end } = state.doc.lineAt(range.from);;) {
            if (start != lastLine) {
                lastLine = start;
                let indent = getIndentation(context, start), current;
                if (indent > -1 &&
                    indent != (current = /^\s*/.exec(state.doc.slice(start, Math.min(end, start + 100)))[0].length)) {
                    updated[start] = indent;
                    positions.push({ pos: start, current, indent });
                }
            }
            if (end + 1 > range.to)
                break;
            ({ start, end } = state.doc.lineAt(end + 1));
        }
    }
    if (positions.length > 0) {
        let tr = state.t();
        for (let { pos, current, indent } of positions) {
            let start = tr.changes.mapPos(pos);
            tr.replace(start, start + current, space(indent));
        }
        dispatch(tr);
    }
    return true;
};
/// The default keymap for Linux/Windows/non-Mac platforms. Binds the
/// arrows for cursor motion, shift-arrow for selection extension,
/// ctrl-arrows for by-word motion, home/end for line start/end,
/// ctrl-home/end for document start/end, ctrl-a to select all,
/// backspace/delete for deletion, and enter for newline-and-indent.
const pcBaseKeymap = {
    "ArrowLeft": moveCharLeft,
    "ArrowRight": moveCharRight,
    "Shift-ArrowLeft": extendCharLeft,
    "Shift-ArrowRight": extendCharRight,
    "Mod-ArrowLeft": moveWordLeft,
    "Mod-ArrowRight": moveWordRight,
    "Shift-Mod-ArrowLeft": extendWordLeft,
    "Shift-Mod-ArrowRight": extendWordRight,
    "ArrowUp": moveLineUp,
    "ArrowDown": moveLineDown,
    "Shift-ArrowUp": extendLineUp,
    "Shift-ArrowDown": extendLineDown,
    "Home": moveLineStart,
    "End": moveLineEnd,
    "Shift-Home": extendLineStart,
    "Shift-End": extendLineEnd,
    "Mod-Home": selectDocStart,
    "Mod-End": selectDocEnd,
    "Mod-a": selectAll,
    "Backspace": deleteCharBackward,
    "Delete": deleteCharForward,
    "Enter": insertNewlineAndIndent
};
/// The default keymap for Mac platforms. Includes the bindings from
/// the [PC keymap](#commands.pcBaseKeymap) (using Cmd instead of
/// Ctrl), and adds Mac-specific default bindings.
const macBaseKeymap = {
    "Control-b": moveCharLeft,
    "Control-f": moveCharRight,
    "Shift-Control-b": extendCharLeft,
    "Shift-Control-f": extendCharRight,
    "Control-p": moveLineUp,
    "Control-n": moveLineDown,
    "Shift-Control-p": extendLineUp,
    "Shift-Control-n": extendLineDown,
    "Control-a": moveLineStart,
    "Control-e": moveLineEnd,
    "Shift-Control-a": extendLineStart,
    "Shift-Control-e": extendLineEnd,
    "Cmd-ArrowUp": selectDocStart,
    "Cmd-ArrowDown": selectDocEnd,
    "Control-d": deleteCharForward,
    "Control-h": deleteCharBackward
};
for (let key in pcBaseKeymap)
    macBaseKeymap[key] = pcBaseKeymap[key];
const mac = typeof navigator != "undefined" ? /Mac/.test(navigator.platform)
    : typeof os != "undefined" ? os.platform() == "darwin" : false;
/// The default keymap for the current platform.
const baseKeymap = mac ? macBaseKeymap : pcBaseKeymap;

export { baseKeymap, deleteCharBackward, deleteCharForward, extendCharLeft, extendCharRight, extendLineDown, extendLineEnd, extendLineStart, extendLineUp, extendWordLeft, extendWordRight, indentSelection, insertNewlineAndIndent, macBaseKeymap, moveCharLeft, moveCharRight, moveLineDown, moveLineEnd, moveLineStart, moveLineUp, moveWordLeft, moveWordRight, pcBaseKeymap, selectAll, selectDocEnd, selectDocStart };
